@@ -1668,6 +1668,81 @@ mm_get_string_unquoted_from_match_info (GMatchInfo *match_info,
     return str;
 }
 
+/*
+ * The following implementation is taken from glib g_date_time_format_iso8601 code
+ * https://gitlab.gnome.org/GNOME/glib/-/blob/main/glib/gdatetime.c#L3490
+ */
+static gchar *
+date_time_format_iso8601 (GDateTime *dt)
+{
+#if GLIB_CHECK_VERSION (2, 62, 0)
+    return g_date_time_format_iso8601 (dt);
+#else
+    GString          *outstr = NULL;
+    g_autofree gchar *main_date = NULL;
+    gint64            offset = 0;
+
+    main_date = g_date_time_format (dt, "%Y-%m-%dT%H:%M:%S");
+    outstr = g_string_new (main_date);
+
+    /* Timezone. Format it as `%:::z` unless the offset is zero, in which case
+     * we can simply use `Z`. */
+    offset = g_date_time_get_utc_offset (dt);
+    if (offset == 0) {
+        g_string_append_c (outstr, 'Z');
+    } else {
+        g_autofree gchar *time_zone = NULL;
+
+        time_zone = g_date_time_format (dt, "%:::z");
+        g_string_append (outstr, time_zone);
+    }
+
+    return g_string_free (outstr, FALSE);
+#endif
+}
+
+gchar *
+mm_new_iso8601_time_from_unix_time (guint64 timestamp)
+{
+    g_autoptr(GDateTime) dt = NULL;
+
+    dt = g_date_time_new_from_unix_utc ((gint64)timestamp);
+
+    return date_time_format_iso8601 (dt);
+}
+
+gchar *
+mm_new_iso8601_time (guint    year,
+                     guint    month,
+                     guint    day,
+                     guint    hour,
+                     guint    minute,
+                     guint    second,
+                     gboolean have_offset,
+                     gint     offset_minutes,
+                     GError **error)
+{
+    g_autoptr(GDateTime) dt = NULL;
+
+    if (have_offset) {
+        g_autoptr(GTimeZone) tz = NULL;
+
+        tz = g_time_zone_new_offset (offset_minutes * 60);
+        dt = g_date_time_new (tz, year, month, day, hour, minute, second);
+    } else
+        dt = g_date_time_new_utc (year, month, day, hour, minute, second);
+
+    if (dt == NULL) {
+        g_set_error (error,
+                     MM_CORE_ERROR,
+                     MM_CORE_ERROR_INVALID_ARGS,
+                     "Invalid input for date: got year:%u, month:%u, day:%u, hour:%u, minute:%u, second:%u",
+                     year, month, day, hour, minute, second);
+        return NULL;
+    }
+    return date_time_format_iso8601 (dt);
+}
+
 /*****************************************************************************/
 
 /* From hostap, Copyright (c) 2002-2005, Jouni Malinen <jkmaline@cc.hut.fi> */
