@@ -17,13 +17,17 @@
 #include "mm-modem-helpers.h"
 #include "mm-enums-types.h"
 #include "mm-errors-types.h"
+#include "mm-error-helpers.h"
 #include "mm-log-object.h"
 
+#include <string.h>
+
 /*****************************************************************************/
 
 MMModemCapability
-mm_modem_capability_from_mbim_device_caps (MbimCellularClass caps_cellular_class,
-                                           MbimDataClass     caps_data_class)
+mm_modem_capability_from_mbim_device_caps (MbimCellularClass  caps_cellular_class,
+                                           MbimDataClass      caps_data_class,
+                                           const gchar       *caps_custom_data_class)
 {
     MMModemCapability mask = 0;
 
@@ -38,6 +42,12 @@ mm_modem_capability_from_mbim_device_caps (MbimCellularClass caps_cellular_class
     if (caps_data_class & MBIM_DATA_CLASS_LTE)
         mask |= MM_MODEM_CAPABILITY_LTE;
 
+    if ((caps_data_class & MBIM_DATA_CLASS_CUSTOM) && caps_custom_data_class) {
+        /* e.g. Gosuncn GM800 reports MBIM custom data class "5G/TDS" */
+        if (strstr (caps_custom_data_class, "5G"))
+            mask |= MM_MODEM_CAPABILITY_5GNR;
+    }
+
     return mask;
 }
 
@@ -154,20 +164,23 @@ mm_modem_access_technology_from_mbim_data_class (MbimDataClass data_class)
 MMModem3gppNetworkAvailability
 mm_modem_3gpp_network_availability_from_mbim_provider_state (MbimProviderState state)
 {
-    switch (state) {
-    case MBIM_PROVIDER_STATE_HOME:
-    case MBIM_PROVIDER_STATE_PREFERRED:
-    case MBIM_PROVIDER_STATE_VISIBLE:
-    case MBIM_PROVIDER_STATE_PREFERRED_MULTICARRIER:
-        return MM_MODEM_3GPP_NETWORK_AVAILABILITY_AVAILABLE;
-    case MBIM_PROVIDER_STATE_REGISTERED:
+    /* MbimProviderState is a bitmask!
+     *
+     * We don't explicitly process MBIM_PROVIDER_STATE_PREFERRED,
+     * MBIM_PROVIDER_STATE_PREFERRED_MULTICARRIER or MBIM_PROVIDER_STATE_HOME,
+     * so we don't report at MM level the type of operator it is (home,
+     * preferred or non-preferred), just its availability.
+     */
+    if (state & MBIM_PROVIDER_STATE_REGISTERED)
         return MM_MODEM_3GPP_NETWORK_AVAILABILITY_CURRENT;
-    case MBIM_PROVIDER_STATE_FORBIDDEN:
+
+    if (state & MBIM_PROVIDER_STATE_FORBIDDEN)
         return MM_MODEM_3GPP_NETWORK_AVAILABILITY_FORBIDDEN;
-    case MBIM_PROVIDER_STATE_UNKNOWN:
-    default:
-        return MM_MODEM_3GPP_NETWORK_AVAILABILITY_UNKNOWN;
-    }
+
+    if (state & MBIM_PROVIDER_STATE_VISIBLE)
+        return MM_MODEM_3GPP_NETWORK_AVAILABILITY_AVAILABLE;
+
+    return MM_MODEM_3GPP_NETWORK_AVAILABILITY_UNKNOWN;
 }
 
 /*****************************************************************************/
@@ -198,156 +211,115 @@ mm_3gpp_network_info_list_from_mbim_providers (const MbimProvider *const *provid
 
 /*****************************************************************************/
 
-GError *
-mm_mobile_equipment_error_from_mbim_nw_error (MbimNwError nw_error)
+MbimPinType
+mbim_pin_type_from_mm_modem_3gpp_facility (MMModem3gppFacility facility)
 {
-    const gchar *msg;
-
-    switch (nw_error) {
-    case MBIM_NW_ERROR_IMSI_UNKNOWN_IN_HLR:
-        return g_error_new (MM_MOBILE_EQUIPMENT_ERROR,
-                            MM_MOBILE_EQUIPMENT_ERROR_GPRS_IMSI_UNKNOWN_IN_HLR,
-                            "IMSI unknown in HLR");
-    case MBIM_NW_ERROR_ILLEGAL_MS:
-        return g_error_new (MM_MOBILE_EQUIPMENT_ERROR,
-                            MM_MOBILE_EQUIPMENT_ERROR_GPRS_ILLEGAL_MS,
-                            "Illegal MS");
-    case MBIM_NW_ERROR_IMSI_UNKNOWN_IN_VLR:
-        return g_error_new (MM_MOBILE_EQUIPMENT_ERROR,
-                            MM_MOBILE_EQUIPMENT_ERROR_GPRS_IMSI_UNKNOWN_IN_VLR,
-                            "IMSI unknown in VLR");
-    case MBIM_NW_ERROR_ILLEGAL_ME:
-        return g_error_new (MM_MOBILE_EQUIPMENT_ERROR,
-                            MM_MOBILE_EQUIPMENT_ERROR_GPRS_ILLEGAL_ME,
-                            "Illegal ME");
-    case MBIM_NW_ERROR_GPRS_NOT_ALLOWED:
-        return g_error_new (MM_MOBILE_EQUIPMENT_ERROR,
-                            MM_MOBILE_EQUIPMENT_ERROR_GPRS_SERVICE_NOT_ALLOWED,
-                            "GPRS not allowed");
-    case MBIM_NW_ERROR_GPRS_AND_NON_GPRS_NOT_ALLOWED:
-        return g_error_new (MM_MOBILE_EQUIPMENT_ERROR,
-                            MM_MOBILE_EQUIPMENT_ERROR_GPRS_SERVICE_NOT_ALLOWED,
-                            "GPRS and non-GPRS not allowed");
-    case MBIM_NW_ERROR_PLMN_NOT_ALLOWED:
-        return g_error_new (MM_MOBILE_EQUIPMENT_ERROR,
-                            MM_MOBILE_EQUIPMENT_ERROR_GPRS_PLMN_NOT_ALLOWED,
-                            "PLMN not allowed");
-    case MBIM_NW_ERROR_LOCATION_AREA_NOT_ALLOWED:
-        return g_error_new (MM_MOBILE_EQUIPMENT_ERROR,
-                            MM_MOBILE_EQUIPMENT_ERROR_GPRS_LOCATION_NOT_ALLOWED,
-                            "Location area not allowed");
-    case MBIM_NW_ERROR_ROAMING_NOT_ALLOWED_IN_LOCATION_AREA:
-        return g_error_new (MM_MOBILE_EQUIPMENT_ERROR,
-                            MM_MOBILE_EQUIPMENT_ERROR_GPRS_ROAMING_NOT_ALLOWED,
-                            "Roaming not allowed in location area");
-    case MBIM_NW_ERROR_GPRS_NOT_ALLOWED_IN_PLMN:
-        return g_error_new (MM_MOBILE_EQUIPMENT_ERROR,
-                            MM_MOBILE_EQUIPMENT_ERROR_GPRS_SERVICE_NOT_ALLOWED,
-                            "GPRS not allowed in PLMN");
-    case MBIM_NW_ERROR_NO_CELLS_IN_LOCATION_AREA:
-        return g_error_new (MM_MOBILE_EQUIPMENT_ERROR,
-                            MM_MOBILE_EQUIPMENT_ERROR_GPRS_NO_CELLS_IN_LOCATION_AREA,
-                            "No cells in location area");
-    case MBIM_NW_ERROR_NETWORK_FAILURE:
-        return g_error_new (MM_MOBILE_EQUIPMENT_ERROR,
-                            MM_MOBILE_EQUIPMENT_ERROR_GPRS_NETWORK_FAILURE,
-                            "Network failure");
-    case MBIM_NW_ERROR_CONGESTION:
-        return g_error_new (MM_MOBILE_EQUIPMENT_ERROR,
-                            MM_MOBILE_EQUIPMENT_ERROR_GPRS_CONGESTION,
-                            "Congestion");
-    case MBIM_NW_ERROR_GSM_AUTHENTICATION_UNACCEPTABLE:
-        return g_error_new (MM_MOBILE_EQUIPMENT_ERROR,
-                            MM_MOBILE_EQUIPMENT_ERROR_GPRS_USER_AUTHENTICATION_FAILED,
-                            "GSM authentication unacceptable");
-    case MBIM_NW_ERROR_NOT_AUTHORIZED_FOR_CSG:
-        return g_error_new (MM_MOBILE_EQUIPMENT_ERROR,
-                            MM_MOBILE_EQUIPMENT_ERROR_GPRS_USER_AUTHENTICATION_FAILED,
-                            "Not authorized for this CSG");
-    case MBIM_NW_ERROR_INSUFFICIENT_RESOURCES:
-        return g_error_new (MM_MOBILE_EQUIPMENT_ERROR,
-                            MM_MOBILE_EQUIPMENT_ERROR_GPRS_INSUFFICIENT_RESOURCES,
-                            "Insufficient resources");
-    case MBIM_NW_ERROR_MISSING_OR_UNKNOWN_APN:
-        return g_error_new (MM_MOBILE_EQUIPMENT_ERROR,
-                            MM_MOBILE_EQUIPMENT_ERROR_GPRS_MISSING_OR_UNKNOWN_APN,
-                            "Missing or unknown APN");
-    case MBIM_NW_ERROR_UNKNOWN_PDP_ADDRESS_OR_TYPE:
-        return g_error_new (MM_MOBILE_EQUIPMENT_ERROR,
-                            MM_MOBILE_EQUIPMENT_ERROR_GPRS_UNKNOWN_PDP_ADDRESS_OR_TYPE,
-                            "Unknown PDP address or type");
-    case MBIM_NW_ERROR_USER_AUTHENTICATION_FAILED:
-        return g_error_new (MM_MOBILE_EQUIPMENT_ERROR,
-                            MM_MOBILE_EQUIPMENT_ERROR_GPRS_USER_AUTHENTICATION_FAILED,
-                            "User authentication failed");
-    case MBIM_NW_ERROR_ACTIVATION_REJECTED_BY_GGSN_OR_GW:
-        return g_error_new (MM_MOBILE_EQUIPMENT_ERROR,
-                            MM_MOBILE_EQUIPMENT_ERROR_GPRS_ACTIVATION_REJECTED_BY_GGSN_OR_GW,
-                            "Activation rejected by GGSN or GW");
-    case MBIM_NW_ERROR_ACTIVATION_REJECTED_UNSPECIFIED:
-        return g_error_new (MM_MOBILE_EQUIPMENT_ERROR,
-                            MM_MOBILE_EQUIPMENT_ERROR_GPRS_ACTIVATION_REJECTED_UNSPECIFIED,
-                            "Activation rejected; unspecified");
-    case MBIM_NW_ERROR_SERVICE_OPTION_NOT_SUPPORTED:
-        return g_error_new (MM_MOBILE_EQUIPMENT_ERROR,
-                            MM_MOBILE_EQUIPMENT_ERROR_GPRS_SERVICE_OPTION_NOT_SUPPORTED,
-                            "Service option not supported");
-    case MBIM_NW_ERROR_REQUESTED_SERVICE_OPTION_NOT_SUBSCRIBED:
-        return g_error_new (MM_MOBILE_EQUIPMENT_ERROR,
-                            MM_MOBILE_EQUIPMENT_ERROR_GPRS_SERVICE_OPTION_NOT_SUBSCRIBED,
-                            "Requested service option not subscribed");
-    case MBIM_NW_ERROR_SERVICE_OPTION_TEMPORARILY_OUT_OF_ORDER:
-        return g_error_new (MM_MOBILE_EQUIPMENT_ERROR,
-                            MM_MOBILE_EQUIPMENT_ERROR_GPRS_SERVICE_OPTION_OUT_OF_ORDER,
-                            "Service option temporarily out of order");
-    case MBIM_NW_ERROR_MAXIMUM_NUMBER_OF_PDP_CONTEXTS_REACHED:
-        return g_error_new (MM_MOBILE_EQUIPMENT_ERROR,
-                            MM_MOBILE_EQUIPMENT_ERROR_GPRS_MAXIMUM_NUMBER_OF_PDP_CONTEXTS_REACHED,
-                            "Maximum number of PDP contexts reached");
-    case MBIM_NW_ERROR_REQUESTED_APN_NOT_SUPPORTED_IN_CURRENT_RAT_AND_PLMN:
-        return g_error_new (MM_MOBILE_EQUIPMENT_ERROR,
-                            MM_MOBILE_EQUIPMENT_ERROR_GPRS_REQUESTED_APN_NOT_SUPPORTED,
-                            "Requested APN not supported");
-
-    case MBIM_NW_ERROR_SEMANTICALLY_INCORRECT_MESSAGE:
-        return g_error_new (MM_MOBILE_EQUIPMENT_ERROR,
-                            MM_MOBILE_EQUIPMENT_ERROR_GPRS_SEMANTICALLY_INCORRECT_MESSAGE,
-                            "Semantically incorrect message");
-
-    case MBIM_NW_ERROR_PROTOCOL_ERROR_UNSPECIFIED:
-        return g_error_new (MM_MOBILE_EQUIPMENT_ERROR,
-                            MM_MOBILE_EQUIPMENT_ERROR_GPRS_UNSPECIFIED_PROTOCOL_ERROR,
-                            "Unspecified protocol error");
-
-    case MBIM_NW_ERROR_IMEI_NOT_ACCEPTED:
-    case MBIM_NW_ERROR_MS_IDENTITY_NOT_DERIVED_BY_NETWORK:
-    case MBIM_NW_ERROR_IMPLICITLY_DETACHED:
-    case MBIM_NW_ERROR_MSC_TEMPORARILY_NOT_REACHABLE:
-    case MBIM_NW_ERROR_MAC_FAILURE:
-    case MBIM_NW_ERROR_SYNCH_FAILURE:
-    case MBIM_NW_ERROR_NO_PDP_CONTEXT_ACTIVATED:
-    case MBIM_NW_ERROR_PDP_TYPE_IPV4_ONLY_ALLOWED:
-    case MBIM_NW_ERROR_PDP_TYPE_IPV6_ONLY_ALLOWED:
-    case MBIM_NW_ERROR_INVALID_MANDATORY_INFORMATION:
-    case MBIM_NW_ERROR_MESSAGE_TYPE_NON_EXISTENT_OR_NOT_IMPLEMENTED:
-    case MBIM_NW_ERROR_MESSAGE_TYPE_NOT_COMPATIBLE_WITH_PROTOCOL_STATE:
-    case MBIM_NW_ERROR_INFORMATION_ELEMENT_NON_EXISTENT_OR_NOT_IMPLEMENTED:
-    case MBIM_NW_ERROR_CONDITIONAL_IE_ERROR:
-    case MBIM_NW_ERROR_MESSAGE_NOT_COMPATIBLE_WITH_PROTOCOL_STATE:
-    case MBIM_NW_ERROR_APN_RESTRICTION_VALUE_INCOMPATIBLE_WITH_ACTIVE_PDP_CONTEXT:
-    case MBIM_NW_ERROR_MULTIPLE_ACCESSES_TO_A_PDN_CONNECTION_NOT_ALLOWED:
-    case MBIM_NW_ERROR_UNKNOWN:
+    switch (facility) {
+    case MM_MODEM_3GPP_FACILITY_NET_PERS:
+        return MBIM_PIN_TYPE_NETWORK_PIN;
+    case MM_MODEM_3GPP_FACILITY_NET_SUB_PERS:
+        return MBIM_PIN_TYPE_NETWORK_SUBSET_PIN;
+    case MM_MODEM_3GPP_FACILITY_PROVIDER_PERS:
+        return MBIM_PIN_TYPE_SERVICE_PROVIDER_PIN;
+    case MM_MODEM_3GPP_FACILITY_CORP_PERS:
+        return MBIM_PIN_TYPE_CORPORATE_PIN;
+    case MM_MODEM_3GPP_FACILITY_SIM:
+        return MBIM_PIN_TYPE_PIN1;
+    case MM_MODEM_3GPP_FACILITY_FIXED_DIALING:
+        return MBIM_PIN_TYPE_PIN2;
+    case MM_MODEM_3GPP_FACILITY_PH_SIM:
+        return MBIM_PIN_TYPE_DEVICE_SIM_PIN;
+    case MM_MODEM_3GPP_FACILITY_PH_FSIM:
+        return MBIM_PIN_TYPE_DEVICE_FIRST_SIM_PIN;
+    case MM_MODEM_3GPP_FACILITY_NONE:
     default:
-        msg = mbim_nw_error_get_string (nw_error);
-        return g_error_new (MM_MOBILE_EQUIPMENT_ERROR,
-                            MM_MOBILE_EQUIPMENT_ERROR_GPRS_UNKNOWN,
-                            "Unsupported error (%u): %s",
-                            nw_error, msg ? msg : "unknown");
+        return MBIM_PIN_TYPE_UNKNOWN;
     }
 }
 
 /*****************************************************************************/
 
+static const MMMobileEquipmentError mbim_nw_errors[] = {
+    [MBIM_NW_ERROR_IMSI_UNKNOWN_IN_HLR] = MM_MOBILE_EQUIPMENT_ERROR_IMSI_UNKNOWN_IN_HSS,
+    [MBIM_NW_ERROR_ILLEGAL_MS] = MM_MOBILE_EQUIPMENT_ERROR_ILLEGAL_UE,
+    [MBIM_NW_ERROR_IMSI_UNKNOWN_IN_VLR] = MM_MOBILE_EQUIPMENT_ERROR_IMSI_UNKNOWN_IN_VLR,
+    [MBIM_NW_ERROR_ILLEGAL_ME] = MM_MOBILE_EQUIPMENT_ERROR_ILLEGAL_ME,
+    [MBIM_NW_ERROR_GPRS_NOT_ALLOWED] = MM_MOBILE_EQUIPMENT_ERROR_PS_SERVICES_NOT_ALLOWED,
+    [MBIM_NW_ERROR_GPRS_AND_NON_GPRS_NOT_ALLOWED] = MM_MOBILE_EQUIPMENT_ERROR_PS_AND_NON_PS_SERVICES_NOT_ALLOWED,
+    [MBIM_NW_ERROR_PLMN_NOT_ALLOWED] = MM_MOBILE_EQUIPMENT_ERROR_PLMN_NOT_ALLOWED,
+    [MBIM_NW_ERROR_LOCATION_AREA_NOT_ALLOWED] = MM_MOBILE_EQUIPMENT_ERROR_AREA_NOT_ALLOWED,
+    [MBIM_NW_ERROR_ROAMING_NOT_ALLOWED_IN_LOCATION_AREA] = MM_MOBILE_EQUIPMENT_ERROR_ROAMING_NOT_ALLOWED_IN_AREA,
+    [MBIM_NW_ERROR_GPRS_NOT_ALLOWED_IN_PLMN] = MM_MOBILE_EQUIPMENT_ERROR_PS_SERVICES_NOT_ALLOWED_IN_PLMN,
+    [MBIM_NW_ERROR_NO_CELLS_IN_LOCATION_AREA] = MM_MOBILE_EQUIPMENT_ERROR_NO_CELLS_IN_AREA,
+    [MBIM_NW_ERROR_NETWORK_FAILURE] = MM_MOBILE_EQUIPMENT_ERROR_NETWORK_FAILURE_ATTACH,
+    [MBIM_NW_ERROR_CONGESTION] = MM_MOBILE_EQUIPMENT_ERROR_CONGESTION,
+    [MBIM_NW_ERROR_GSM_AUTHENTICATION_UNACCEPTABLE] = MM_MOBILE_EQUIPMENT_ERROR_USER_AUTHENTICATION_FAILED,
+    [MBIM_NW_ERROR_NOT_AUTHORIZED_FOR_CSG] = MM_MOBILE_EQUIPMENT_ERROR_NOT_AUTHORIZED_FOR_CSG,
+    [MBIM_NW_ERROR_INSUFFICIENT_RESOURCES] = MM_MOBILE_EQUIPMENT_ERROR_INSUFFICIENT_RESOURCES,
+    [MBIM_NW_ERROR_MISSING_OR_UNKNOWN_APN] = MM_MOBILE_EQUIPMENT_ERROR_MISSING_OR_UNKNOWN_APN,
+    [MBIM_NW_ERROR_UNKNOWN_PDP_ADDRESS_OR_TYPE] = MM_MOBILE_EQUIPMENT_ERROR_UNKNOWN_PDP_ADDRESS_OR_TYPE,
+    [MBIM_NW_ERROR_USER_AUTHENTICATION_FAILED] = MM_MOBILE_EQUIPMENT_ERROR_USER_AUTHENTICATION_FAILED,
+    [MBIM_NW_ERROR_ACTIVATION_REJECTED_BY_GGSN_OR_GW] = MM_MOBILE_EQUIPMENT_ERROR_ACTIVATION_REJECTED_BY_GGSN_OR_GW,
+    [MBIM_NW_ERROR_ACTIVATION_REJECTED_UNSPECIFIED] = MM_MOBILE_EQUIPMENT_ERROR_ACTIVATION_REJECTED_UNSPECIFIED,
+    [MBIM_NW_ERROR_SERVICE_OPTION_NOT_SUPPORTED] = MM_MOBILE_EQUIPMENT_ERROR_SERVICE_OPTION_NOT_SUPPORTED,
+    [MBIM_NW_ERROR_REQUESTED_SERVICE_OPTION_NOT_SUBSCRIBED] = MM_MOBILE_EQUIPMENT_ERROR_SERVICE_OPTION_NOT_SUBSCRIBED,
+    [MBIM_NW_ERROR_SERVICE_OPTION_TEMPORARILY_OUT_OF_ORDER] = MM_MOBILE_EQUIPMENT_ERROR_SERVICE_OPTION_OUT_OF_ORDER,
+    [MBIM_NW_ERROR_MAXIMUM_NUMBER_OF_PDP_CONTEXTS_REACHED] = MM_MOBILE_EQUIPMENT_ERROR_MAXIMUM_NUMBER_OF_BEARERS_REACHED,
+    [MBIM_NW_ERROR_REQUESTED_APN_NOT_SUPPORTED_IN_CURRENT_RAT_AND_PLMN] = MM_MOBILE_EQUIPMENT_ERROR_REQUESTED_APN_NOT_SUPPORTED,
+    [MBIM_NW_ERROR_SEMANTICALLY_INCORRECT_MESSAGE] = MM_MOBILE_EQUIPMENT_ERROR_SEMANTICALLY_INCORRECT_MESSAGE,
+    [MBIM_NW_ERROR_PROTOCOL_ERROR_UNSPECIFIED] = MM_MOBILE_EQUIPMENT_ERROR_UNSPECIFIED_PROTOCOL_ERROR,
+    [MBIM_NW_ERROR_IMEI_NOT_ACCEPTED] = MM_MOBILE_EQUIPMENT_ERROR_IMEI_NOT_ACCEPTED,
+    [MBIM_NW_ERROR_MS_IDENTITY_NOT_DERIVED_BY_NETWORK] = MM_MOBILE_EQUIPMENT_ERROR_UE_IDENTITY_NOT_DERIVED_FROM_NETWORK,
+    [MBIM_NW_ERROR_IMPLICITLY_DETACHED] = MM_MOBILE_EQUIPMENT_ERROR_IMPLICITLY_DETACHED,
+    [MBIM_NW_ERROR_MSC_TEMPORARILY_NOT_REACHABLE] = MM_MOBILE_EQUIPMENT_ERROR_MSC_TEMPORARILY_NOT_REACHABLE,
+    [MBIM_NW_ERROR_NO_PDP_CONTEXT_ACTIVATED] = MM_MOBILE_EQUIPMENT_ERROR_NO_BEARER_ACTIVATED,
+    [MBIM_NW_ERROR_PDP_TYPE_IPV4_ONLY_ALLOWED] = MM_MOBILE_EQUIPMENT_ERROR_IPV4_ONLY_ALLOWED,
+    [MBIM_NW_ERROR_PDP_TYPE_IPV6_ONLY_ALLOWED] = MM_MOBILE_EQUIPMENT_ERROR_IPV6_ONLY_ALLOWED,
+    [MBIM_NW_ERROR_INVALID_MANDATORY_INFORMATION] = MM_MOBILE_EQUIPMENT_ERROR_INVALID_MANDATORY_INFORMATION,
+    [MBIM_NW_ERROR_MESSAGE_TYPE_NON_EXISTENT_OR_NOT_IMPLEMENTED] = MM_MOBILE_EQUIPMENT_ERROR_MESSAGE_TYPE_NOT_IMPLEMENTED,
+    [MBIM_NW_ERROR_MESSAGE_TYPE_NOT_COMPATIBLE_WITH_PROTOCOL_STATE] = MM_MOBILE_EQUIPMENT_ERROR_MESSAGE_TYPE_NOT_COMPATIBLE_WITH_PROTOCOL_STATE,
+    [MBIM_NW_ERROR_INFORMATION_ELEMENT_NON_EXISTENT_OR_NOT_IMPLEMENTED] = MM_MOBILE_EQUIPMENT_ERROR_IE_NOT_IMPLEMENTED,
+    [MBIM_NW_ERROR_CONDITIONAL_IE_ERROR] = MM_MOBILE_EQUIPMENT_ERROR_CONDITIONAL_IE_ERROR,
+    [MBIM_NW_ERROR_MESSAGE_NOT_COMPATIBLE_WITH_PROTOCOL_STATE] = MM_MOBILE_EQUIPMENT_ERROR_MESSAGE_NOT_COMPATIBLE_WITH_PROTOCOL_STATE,
+    [MBIM_NW_ERROR_APN_RESTRICTION_VALUE_INCOMPATIBLE_WITH_ACTIVE_PDP_CONTEXT] = MM_MOBILE_EQUIPMENT_ERROR_APN_RESTRICTION_INCOMPATIBLE,
+    [MBIM_NW_ERROR_MULTIPLE_ACCESSES_TO_A_PDN_CONNECTION_NOT_ALLOWED] = MM_MOBILE_EQUIPMENT_ERROR_MULTIPLE_ACCESS_TO_PDN_CONNECTION_NOT_ALLOWED,
+    [MBIM_NW_ERROR_UNKNOWN] = MM_MOBILE_EQUIPMENT_ERROR_UNKNOWN,
+    /* known unmapped errors */
+    /* MBIM_NW_ERROR_MAC_FAILURE */
+    /* MBIM_NW_ERROR_SYNCH_FAILURE */
+};
+
+GError *
+mm_mobile_equipment_error_from_mbim_nw_error (MbimNwError nw_error,
+                                              gpointer    log_object)
+{
+    const gchar            *msg;
+
+    if (nw_error < G_N_ELEMENTS (mbim_nw_errors)) {
+        MMMobileEquipmentError  error_code;
+
+        /* convert to mobile equipment error */
+        error_code = mbim_nw_errors[nw_error];
+        if (error_code)
+            return mm_mobile_equipment_error_for_code (error_code, log_object);
+
+        /* provide a nicer error message on unmapped errors */
+        msg = mbim_nw_error_get_string (nw_error);
+        if (msg)
+            return g_error_new (MM_MOBILE_EQUIPMENT_ERROR,
+                                MM_MOBILE_EQUIPMENT_ERROR_UNKNOWN,
+                                "Unsupported error (%u): %s",
+                                nw_error, msg);
+    }
+
+    /* fallback */
+    return g_error_new_literal (MM_MOBILE_EQUIPMENT_ERROR,
+                                MM_MOBILE_EQUIPMENT_ERROR_UNKNOWN,
+                                "Unknown error");
+}
+
+/*****************************************************************************/
+
 MMBearerAllowedAuth
 mm_bearer_allowed_auth_from_mbim_auth_protocol (MbimAuthProtocol auth_protocol)
 {
@@ -375,13 +347,13 @@ mm_bearer_allowed_auth_to_mbim_auth_protocol (MMBearerAllowedAuth   bearer_auth,
     /* NOTE: the input is a BITMASK, so we try to find a "best match" */
 
     if (bearer_auth == MM_BEARER_ALLOWED_AUTH_UNKNOWN) {
-        mm_obj_dbg (log_object, "using default (PAP) authentication method");
-        return MBIM_AUTH_PROTOCOL_PAP;
+        mm_obj_dbg (log_object, "using default (CHAP) authentication method");
+        return MBIM_AUTH_PROTOCOL_CHAP;
     }
-    if (bearer_auth & MM_BEARER_ALLOWED_AUTH_PAP)
-        return MBIM_AUTH_PROTOCOL_PAP;
     if (bearer_auth & MM_BEARER_ALLOWED_AUTH_CHAP)
         return MBIM_AUTH_PROTOCOL_CHAP;
+    if (bearer_auth & MM_BEARER_ALLOWED_AUTH_PAP)
+        return MBIM_AUTH_PROTOCOL_PAP;
     if (bearer_auth & MM_BEARER_ALLOWED_AUTH_MSCHAPV2)
         return MBIM_AUTH_PROTOCOL_MSCHAPV2;
     if (bearer_auth & MM_BEARER_ALLOWED_AUTH_NONE)
@@ -399,6 +371,70 @@ mm_bearer_allowed_auth_to_mbim_auth_protocol (MMBearerAllowedAuth   bearer_auth,
 
 /*****************************************************************************/
 
+MMBearerApnType
+mm_bearer_apn_type_from_mbim_context_type (MbimContextType context_type)
+{
+    switch (context_type) {
+        case MBIM_CONTEXT_TYPE_INTERNET:
+            return MM_BEARER_APN_TYPE_DEFAULT;
+        case MBIM_CONTEXT_TYPE_VPN:
+            return MM_BEARER_APN_TYPE_PRIVATE;
+        case MBIM_CONTEXT_TYPE_VOICE:
+            return MM_BEARER_APN_TYPE_VOICE;
+        case MBIM_CONTEXT_TYPE_PURCHASE:
+            return MM_BEARER_APN_TYPE_MANAGEMENT;
+        case MBIM_CONTEXT_TYPE_IMS:
+            return MM_BEARER_APN_TYPE_IMS;
+        case MBIM_CONTEXT_TYPE_MMS:
+            return MM_BEARER_APN_TYPE_MMS;
+        case MBIM_CONTEXT_TYPE_INVALID:
+        case MBIM_CONTEXT_TYPE_NONE:
+        case MBIM_CONTEXT_TYPE_LOCAL:
+        case MBIM_CONTEXT_TYPE_VIDEO_SHARE:
+            /* some types unused right now */
+        default:
+            return MM_BEARER_APN_TYPE_NONE;
+    }
+}
+
+MbimContextType
+mm_bearer_apn_type_to_mbim_context_type (MMBearerApnType   apn_type,
+                                         gpointer          log_object,
+                                         GError          **error)
+{
+    g_autofree gchar *str = NULL;
+
+    /* NOTE: the input is a BITMASK, so we try to find a "best match" */
+
+    if (apn_type == MM_BEARER_APN_TYPE_NONE) {
+        mm_obj_dbg (log_object, "using default (internet) APN type");
+        return MBIM_CONTEXT_TYPE_INTERNET;
+    }
+
+    if (apn_type & MM_BEARER_APN_TYPE_DEFAULT)
+        return MBIM_CONTEXT_TYPE_INTERNET;
+    if (apn_type & MM_BEARER_APN_TYPE_IMS)
+        return MBIM_CONTEXT_TYPE_IMS;
+    if (apn_type & MM_BEARER_APN_TYPE_MMS)
+        return MBIM_CONTEXT_TYPE_MMS;
+    if (apn_type &MM_BEARER_APN_TYPE_MANAGEMENT)
+        return MBIM_CONTEXT_TYPE_PURCHASE;
+    if (apn_type & MM_BEARER_APN_TYPE_VOICE)
+        return MBIM_CONTEXT_TYPE_VOICE;
+    if (apn_type & MM_BEARER_APN_TYPE_PRIVATE)
+        return MBIM_CONTEXT_TYPE_VPN;
+
+    str = mm_bearer_apn_type_build_string_from_mask (apn_type);
+    g_set_error (error,
+                 MM_CORE_ERROR,
+                 MM_CORE_ERROR_UNSUPPORTED,
+                 "Unsupported APN types (%s)",
+                 str);
+    return MBIM_CONTEXT_TYPE_NONE;
+}
+
+/*****************************************************************************/
+
 MMBearerIpFamily
 mm_bearer_ip_family_from_mbim_context_ip_type (MbimContextIpType ip_type)
 {
