@@ -71,15 +71,6 @@ guint mm_netmask_to_cidr (const gchar *netmask);
 GArray *mm_filter_current_bands (const GArray *supported_bands,
                                  const GArray *current_bands);
 
-gchar *mm_new_iso8601_time (guint year,
-                            guint month,
-                            guint day,
-                            guint hour,
-                            guint minute,
-                            guint second,
-                            gboolean have_offset,
-                            gint offset_minutes);
-
 GArray *mm_filter_supported_modes (const GArray *all,
                                    const GArray *supported_combinations,
                                    gpointer      log_object);
@@ -169,6 +160,7 @@ gboolean mm_3gpp_parse_cops_read_response (const gchar              *response,
                                            guint                    *out_format,
                                            gchar                   **out_operator,
                                            MMModemAccessTechnology  *out_act,
+                                           gpointer                  log_object,
                                            GError                  **error);
 
 /* Logic to compare two APN names */
